import inspect
from functools import wraps
from typing import Any, Callable, Dict, Tuple

import requests

from abstra_internals.credentials import get_credentials
from abstra_internals.environment import CLOUD_API_CLI_URL
from abstra_internals.threaded import threaded
from abstra_internals.utils import (
    get_local_python_version,
    get_local_user_id,
    is_dev_env,
    is_test_env,
)
from abstra_internals.utils.packages import get_local_package_version


@threaded
def send_usage(payload: Dict):
    if is_test_env() or is_dev_env():
        return

    data = {
        "payload": payload,
        "userId": get_local_user_id(),
        "pythonVersion": get_local_python_version(),
        "abstraVersion": str(get_local_package_version()),
    }

    headers = {"apiKey": get_credentials()}
    api_url = f"{CLOUD_API_CLI_URL}/editor/usage"
    requests.post(api_url, json=data, headers=headers)


def usage(func: Callable[..., Any]) -> Callable[..., Any]:
    @wraps(func)
    def wrapper(*args: Tuple[Any], **kwargs: Any) -> Any:
        arg_names = inspect.getfullargspec(func).args
        arg_values = dict(zip(arg_names, args))

        send_usage({**arg_values, **kwargs, **{"event": func.__name__}})
        return func(*args, **kwargs)

    return wrapper


def track_usage(*, event: str, payload: Dict):
    send_usage({**payload, "event": event})



from typing import Optional

import flask

from abstra_internals.repositories.project.project import ProjectRepository
from abstra_internals.repositories.stage_run import (
    GetStageRunByQueryFilter,
    Pagination,
    StageRunRepository,
)
from abstra_internals.server.controller.main import MainController
from abstra_internals.usage import usage


class StageRunsController:
    def __init__(
        self,
        stage_run_repository: StageRunRepository,
    ) -> None:
        self.stage_run_repository = stage_run_repository

    def get_leaf_stage_runs(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ):
        return [
            stage_run.to_dto()
            for stage_run in self.stage_run_repository.find_leaves(
                filter=filter, pagination=pagination
            ).stage_runs
        ]

    def get_past_stage_runs(
        self, filter: GetStageRunByQueryFilter, pagination: Pagination
    ):
        return [
            stage_run.to_dto()
            for stage_run in self.stage_run_repository.find_past(
                filter=filter, pagination=pagination
            ).stage_runs
        ]

    def fork(self, stage_run_id: str, custom_thread_data: Optional[str]):
        stage_run = self.stage_run_repository.get(stage_run_id)
        project = ProjectRepository.load()
        parent_is_iterator = False
        previous_stages = project.get_previous_stages_ids(stage_run.stage)
        some_is_iterator = False
        for previous_stage in previous_stages:
            previous_stage = project.get_stage(previous_stage)
            if previous_stage and previous_stage.type_name == "iterator":
                some_is_iterator = True
        if some_is_iterator and "item" in stage_run.data:
            parent_is_iterator = True
        return self.stage_run_repository.fork(
            stage_run, parent_is_iterator, custom_thread_data
        ).to_dto()

    def retry(self, stage_run_id: str):
        return self.stage_run_repository.retry(stage_run_id).to_dto()


def get_editor_bp(main_controller: MainController):
    bp = flask.Blueprint("editor_stage_runs", __name__)
    controller = StageRunsController(main_controller.stage_run_repository)

    @bp.get("/")
    @usage
    def _get_leaf_stage_runs():
        filter = GetStageRunByQueryFilter.from_dict(flask.request.args)
        pagination = Pagination.from_dict(flask.request.args)
        return controller.get_leaf_stage_runs(filter, pagination)

    @bp.get("/past")
    @usage
    def _get_past_stage_runs():
        filter = GetStageRunByQueryFilter.from_dict(flask.request.args)
        pagination = Pagination.from_dict(flask.request.args)
        return controller.get_past_stage_runs(filter, pagination)

    @bp.post("/fork")
    @usage
    def _fork():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        custom_thread_data = data.get("custom_thread_data")
        if not stage_run_id:
            flask.abort(400)
        return controller.fork(stage_run_id, custom_thread_data)

    @bp.post("/retry")
    @usage
    def _retry():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        if not stage_run_id:
            flask.abort(400)
        return controller.retry(stage_run_id)

    return bp


def get_player_bp(main_controller: MainController):
    bp = flask.Blueprint("player_stage_runs", __name__)
    controller = StageRunsController(main_controller.stage_run_repository)

    @bp.post("/retry")
    def _retry():
        data = flask.request.json
        if not data:
            flask.abort(400)
        stage_run_id = data.get("stage_run_id")
        if not stage_run_id:
            flask.abort(400)
        return controller.retry(stage_run_id)

    return bp



from abstra_internals.interface.sdk.user_exceptions import MissingDependencyError
from abstra_internals.repositories.connectors import ConnectorsRepository
from abstra_internals.utils.datetime import from_utc_iso_string


def create_abstra_google_credentials(
    repository: ConnectorsRepository, connection_name: str
):
    try:
        from google.auth.credentials import (  # type: ignore
            Credentials as BaseCredentials,
        )
    except ImportError:
        raise MissingDependencyError("google-auth")

    class AbstraGoogleCredentials(BaseCredentials):
        def __init__(
            self, repository: ConnectorsRepository, connection_name: str
        ) -> None:
            super().__init__()
            self.repository = repository
            self.connection_name = connection_name

        def refresh(self, request):
            dto = self.repository.get_access_token(self.connection_name)
            self.token = dto.token
            self.expiry = from_utc_iso_string(dto.expiresAt)
            self.expiry = self.expiry.replace(tzinfo=None)

    return AbstraGoogleCredentials(repository, connection_name)



import json
from abc import ABC, abstractmethod
from dataclasses import dataclass
from datetime import datetime
from pathlib import Path
from typing import Any, Dict, List, Literal, Optional, TypedDict, Union

import requests

from abstra_internals.environment import SIDECAR_HEADERS, SIDECAR_URL
from abstra_internals.logger import AbstraLogger
from abstra_internals.repositories.serializer import SerializationHelper
from abstra_internals.threaded import threaded
from abstra_internals.utils import serialize
from abstra_internals.utils.datetime import from_utc_iso_string, to_utc_iso_string
from abstra_internals.utils.dot_abstra import LOCAL_LOGS_FOLDER


class LogEntryFactory:
    @staticmethod
    def from_dto(dto: Dict) -> "LogEntry":
        if dto.get("event") == "stdout" or dto.get("event") == "stderr":
            return StdioLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                event=dto["event"],
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "form-message":
            return FormEventLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        elif dto.get("event") == "unhandled-exception":
            return UnhandledExceptionLogEntry(
                execution_id=dto["executionId"],
                created_at=from_utc_iso_string(dto["createdAt"]),
                payload=dto["payload"],
                sequence=dto["sequence"],
            )
        else:
            raise Exception("Invalid log entry type")


@dataclass
class StdioLogEntry:
    execution_id: str
    created_at: datetime
    event: Literal["stderr", "stdout"]
    payload: Dict[Literal["text"], str]
    sequence: int

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class FormEventLogEntry:
    execution_id: str
    created_at: datetime
    sequence: int
    payload: Dict[Literal["messageType", "messagePayload"], Union[str, Dict]]
    event: Literal["form-message"] = "form-message"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


@dataclass
class UnhandledExceptionLogEntry:
    execution_id: str
    created_at: datetime
    payload: Dict[Literal["text"], str]
    sequence: int
    event: Literal["unhandled-exception"] = "unhandled-exception"

    def to_dto(self) -> Dict[str, Any]:
        return {
            "executionId": self.execution_id,
            "createdAt": to_utc_iso_string(self.created_at),
            "event": self.event,
            "payload": self.payload,
            "sequence": self.sequence,
        }


LogEntry = Union[StdioLogEntry, FormEventLogEntry, UnhandledExceptionLogEntry]
LogEvent = Literal["stdout", "stderr", "form-message", "unhandled-exception"]
LogsDTO = TypedDict("LogsDTO", {"type": str, "text": str})


class ExecutionLogsRepository(ABC):
    sequence: int

    def get_sequence(self) -> int:
        self.sequence += 1
        return self.sequence

    def insert_stdio(
        self, execution_id: str, event: Literal["stdout", "stderr"], text: str
    ):
        self.save(
            StdioLogEntry(
                execution_id=execution_id,
                created_at=datetime.now(),
                event=event,
                payload={"text": text},
                sequence=self.get_sequence(),
            )
        )

    @abstractmethod
    def save(self, log_entry: LogEntry) -> None:
        raise NotImplementedError()

    @abstractmethod
    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        raise NotImplementedError()

    def get_logs_dto(self, execution_id: str) -> List[LogsDTO]:
        logs = self.get(execution_id)
        output: List[LogsDTO] = []
        for entry in logs:
            if isinstance(entry, FormEventLogEntry):
                continue

            text = entry.payload.get("text")
            if not text:
                continue

            if entry.event == "stdout":
                output.append({"type": "stdout", "text": text})
            elif entry.event == "stderr":
                output.append({"type": "stderr", "text": text})
            elif entry.event == "unhandled-exception":
                output.append({"type": "stderr", "text": text})

        return output


class LocalExecutionLogsRepository(ExecutionLogsRepository):
    _logs: Dict[str, List[LogEntry]]

    def __init__(self):
        self.sequence = 0
        self._logs = {}

    def save(self, log_entry: LogEntry) -> None:
        execution_id = log_entry.execution_id

        if not execution_id:
            return

        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.parent.exists():
            log_file.parent.mkdir(parents=True)

        with open(log_file, "a", encoding="utf-8") as f:
            f.write(serialize(log_entry.to_dto()) + "\n")

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        log_file = Path(LOCAL_LOGS_FOLDER) / f"{execution_id}.log"

        if not log_file.exists():
            return []

        try:
            with open(log_file, "r", encoding="utf-8") as f:
                logs: List[LogEntry] = []
                for line in f.readlines():
                    dto = json.loads(line)
                    if event and dto["event"] != event:
                        continue
                    logs.append(LogEntryFactory.from_dto(dto))
                return logs
        except Exception as e:
            AbstraLogger.capture_exception(e)
            return []


class RemoteExecutionLogsRepository(ExecutionLogsRepository):
    def __init__(self, url: str, headers: Dict[str, str]):
        self.url = url
        self.headers = headers
        self.sequence = 0

    @threaded
    def save(self, log_entry: LogEntry) -> None:
        validated_payload = SerializationHelper.enforce_max_size(log_entry.payload)

        dto = {
            **log_entry.to_dto(),
            "payload": validated_payload,
        }

        res = requests.post(
            f"{self.url}/executions/{log_entry.execution_id}/logs",
            json=dto,
            headers=self.headers,
        )

        res.raise_for_status()

    def get(
        self,
        execution_id: str,
        event: Optional[LogEvent] = None,
    ) -> List[LogEntry]:
        response = requests.get(
            f"{self.url}/executions/{execution_id}/logs",
            params={"event": event} if event else None,
            headers=self.headers,
        )

        response.raise_for_status()

        return [LogEntryFactory.from_dto(log) for log in response.json()]


def execution_logs_repository_factory() -> ExecutionLogsRepository:
    if SIDECAR_URL:
        return RemoteExecutionLogsRepository(url=SIDECAR_URL, headers=SIDECAR_HEADERS)
    else:
        return LocalExecutionLogsRepository()
